#!/usr/bin/env python3
import requests
import socket
import json
import pdb
import subprocess
import platform
from threading import Timer
import urllib3
import os
import sys

try:
    import re
except:
    pass

try:
    import psutil
except:
    pass

try:
    import struct
except:
    pass 

class APIManager():

    checkin_interval = 30
    job_results = []
    tools_folder = os.path.join(os.getcwd(),"tools") + os.sep
    screenconnect_image_path = ""


    def __init__(self, config = []):
        self.config = config

        # disable unverified https warning
        #from requests.packages.urllib3.exceptions import InsecureRequestWarning
        #requests.packages.urllib3.disable_warnings(InsecureRequestWarning)
        
        #urllib3.disable_warnings(urllib3.exceptions.InsecureRequestWarning)
        
        try:
            if not os.path.exists(self.tools_folder):
                os.mkdir(self.tools_folder)
                print('Tools folder created')
        finally:
            pass
    def test(self, url):
        payload =   {
                    'api_action' : 'msp',
                    'api_version' : 1,
                    'task' : 'test'
                    }
        
        try:
            returned = self.post(url, payload)
        except:
            return False
            
        if 'test_success' in returned and returned['test_success'] == 1:
            print('Successful test')
            return True
        else:
            print('Unsuccessful test')
            return False
    
    def create(self, url, hostname):
        payload =   {
                    'api_action' : 'msp',
                    'api_version' : 1,
                    'task' : 'create',
                    'name' : hostname
                    }

        returned = self.post(url, payload)
        if 'success' in returned and returned['success'] == 1:
            print('Successful checkin')
            print(returned)
            return returned
        else:
            print('Unsuccessful checkin')
            print(returned)
            return {'success': 0}

    def getmacbyip(self, ip, pingAddress):

        try:
            if pingAddress:
                Pingparam = '-n' if platform.system().lower()=='windows' else '-c'

                # Building the command. Ex: "ping -c 1 google.com"
                commandPar = ['ping', Pingparam, '1', ip]

                pingResultTemp = subprocess.call(commandPar)

            pid = subprocess.Popen(["arp", "-n", ip], stdout=subprocess.PIPE)
            s = pid.communicate()[0]
            output_text = s.decode('utf-8')
            mac = re.search(r"(([a-f\d]{1,2}\:){5}[a-f\d]{1,2})", output_text).groups()[0]

            return mac.upper()

        except Exception as e: 
            print(e)
            return ''

    def get_default_gateway(self):

        returnInfo = {
            "gateway" : '',
            "interface" : ''
        }

        try:
            """Read the default gateway directly from /proc."""
            with open("/proc/net/route") as fh:
                for line in fh:
                    fields = line.strip().split()
                    if fields[1] != '00000000' or not int(fields[3], 16) & 2:
                        # If not default route or not RTF_GATEWAY, skip it
                        continue

                    returnInfo["gateway"] = socket.inet_ntoa(struct.pack("<L", int(fields[2], 16)))
                    returnInfo["interface"] = fields[0]
        except Exception as e: 
            print(e)

        return returnInfo

    def checkin(self, data):
        '''synchronously checkin to specified server'''
        #print('checkin')
        if self.screenconnect_image_path:
            data['screenconnect_image_path']   = self.screenconnect_image_path
        
        data['job_results'] = self.job_results

        #print(self.job_results)
        payload =   {
                    'api_action' : 'msp',
                    'api_version' : 1,
                    'task' : 'checkin',
                    'uuid' : self.config['uuid'],
                    'data' : json.dumps(data)
                    }
                    
        self.job_results = []

        returned = self.post(self.config['url'], payload)
        
        #print(returned)
        
        if 'success' in returned and returned['success'] == 1:
            #print statement commented out to prevent output file blow up -Lachlan VT
            #print('Successful checkin')
            #pass

            # Get ScreenConnect Info
            try:
                if 'screenconnect_instance_id' in returned:
                    screenconnect_file = "/opt/connectwisecontrol-"+returned['screenconnect_instance_id']+"/ClientLaunchParameters.txt"
                    if (os.path.isfile(screenconnect_file)):
                        with open(screenconnect_file, 'r') as screenconnect_file:
                            self.screenconnect_image_path = screenconnect_file.read()
            except:
                pass

            if 'jobs' in returned:
                if len(returned['jobs']) > 0:
                    for jobsdict in returned['jobs']:
                        #print('command received')
                        if jobsdict['type'] == 'execute_command':
                            #print("the arguments are",jobsdict['arguments'])
                            #for cmd in jobsdict['arguments']:
                            commandsdict=json.loads(jobsdict['arguments'])
                            #the jobs are returned as a string with a dictionary, this line turns it into executeable python code
                            #print('the dictionary is set to',commandsdict)
                            #this works for only one command but will need to be modified for multiple
                            cmd = commandsdict["command"]
                            #print('the command is',cmd.split())
                            #this runs the shell command and is the best way to return the result
                            
                            
                            cmdresult = ''
                            success = 0
                            
                            try:
                            
                                if 'use_shell' in commandsdict and commandsdict['use_shell'] == 1:
                                    try:
                                        subprocess.Popen(cmd.split(),shell=True)
                                    finally:
                                        pass
                                
                                else:
                                
                                    proc = subprocess.Popen(cmd.split(),stdout=subprocess.PIPE)
                                    timer = Timer(60, proc.kill)
                                    try:
                                        timer.start()
                                        cmdresult = proc.communicate()[0]
                                    finally:
                                        timer.cancel()
                                
                                success = 1
                            except:
                                pass
                                
                            try:
                                cmdresult = cmdresult.decode('utf-8')
                            except:
                                pass
                                
                            commandResult = {
                                "id" : jobsdict['id'],
                                "message" : cmdresult,
                                "success" : success
                            }
                            
                            self.job_results.append(commandResult)
                            del commandResult
                            
                            #print(cmdresult)
                        elif jobsdict['type'] == 'download_file_from_url':
                            
                            
                            success = 0
                            message = ''

                            
                            try:
                                downloadDict=json.loads(jobsdict['arguments'])
                                
                                file_name = os.path.join(self.tools_folder, downloadDict['file_name'])
                    
                                if os.path.isfile(file_name) and downloadDict['replace_existing_file'] == 0:
                                    success = 0
                                    message = 'Download Error: File already exists'
                                else:
            
                                    try:
                                    
                                        r = requests.get(downloadDict['url'])
                                    
                                        with open(file_name, 'wb') as f:
                                            f.write(r.content)
                                            
                                        success = 1
                                        message = "Download Success: " + downloadDict['url'] + " To: " + file_name
                                    except requests.exceptions.RequestException as e:
                                        message = str(e)
                                        success= 0
                            except:
                                pass
                                
                            downloadResult = {
                                "id" : jobsdict['id'],
                                "message" : message,
                                "success" : success
                            
                            }
                            
                            self.job_results.append(downloadResult)
                            del downloadResult
                            
                        
                        elif jobsdict['type'] == 'get_processes':
                            processes = []
                             
                            for proc in psutil.process_iter():
                                try:
                                    # Get process name & pid from process object.
                                    
                                    process = {
                                        "name" : proc.name(),
                                        "pid" : proc.pid,
                                        "cpu_time" : proc.cpu_percent(),
                                        "physical_ram" : '',
                                        "peak_ram" : '',
                                        "file_name" : proc.exe()
                                    }
                                    
                                    
                                    processes.append(process)
                                    del process
                                    
                                    
                                except:
                                    pass
                            
                            processResult = {
                                "id" : jobsdict['id'],
                                "message" : '',
                                "success" : 1,
                                "processes" : processes
                            
                            }
                            
                            self.job_results.append(processResult)   
                            del processResult
                        
                        elif jobsdict['type'] == 'get_services':
                            services = []

                            try: 
                                p =  subprocess.Popen(["systemctl", "--type=service", "--all", "--no-pager"], stdout=subprocess.PIPE)
                                (output, err) = p.communicate()
                                output = output.decode('utf-8')
                                
                                service_status = {}
                                for line in output.splitlines():
                                    try:

                                        serviceFullName = ""
                                        serviceStatus = ""
                                        serviceName = ""

                                        if line == "":
                                            continue
                                        
                                        others = line.split()
                                        description = ' '.join(line.split()[4:])
                                        
                                        if others[0] == "●":
                                            description = ' '.join(line.split()[5:])
                                        
                                        if others[0] == "●":
                                            serviceFullName = description
                                            serviceStatus = others[3]
                                            serviceName = others[1]
                                        else:
                                            serviceFullName = description
                                            serviceStatus = others[2]
                                            serviceName = others[0]
                                        

                                        testServiceName =serviceName.endswith(".service")

                                        if (testServiceName):

                                            Servicehead, Servicesep, Servicetail = serviceName.partition('.service')

                                            if (serviceStatus == "active"):
                                                serviceStatus = "Running"
                                            else:
                                                serviceStatus = "Stopped"

                                            service = {
                                                "name" : serviceFullName,
                                                "status" : serviceStatus,
                                                "service" : Servicehead
                                            } 
                                                
                                            services.append(service) 
                                            del(service)                                       
                                    
                                        success = 1

                                    except Exception as error:
                                        print(error)
                                        pass
                            
                            except Exception as error:
                                print(error)
                                success = 0
                            pass
                                
                            serviceResult = {
                                "id" : jobsdict['id'],
                                "message" : '',
                                "success" : success,
                                "services" : services
                            }
                            
                            self.job_results.append(serviceResult)   
                            del serviceResult
                        elif jobsdict['type'] == 'restart_service': 
                            
                            try:
                                restartServiceDict=json.loads(jobsdict['arguments'])

                                serviceName = restartServiceDict['service_name'] + ".service"
                                
                                p =  subprocess.Popen(["systemctl", "restart", serviceName], stdout=subprocess.PIPE)
                                (output, err) = p.communicate()
                                output = output.decode('utf-8')
                                success = 1
                            except:
                                success = 0
                                pass

                            serviceRestartResult = {
                                "id" : jobsdict['id'],
                                "message" : '',
                                "success" : success,
                            }
                            
                            self.job_results.append(serviceRestartResult)   
                            del serviceRestartResult  
                        elif jobsdict['type'] == 'stop_service': 
                            
                            try:
                                restartServiceDict=json.loads(jobsdict['arguments'])

                                serviceName = restartServiceDict['service_name'] + ".service"
                                
                                p =  subprocess.Popen(["systemctl", "stop", serviceName], stdout=subprocess.PIPE)
                                (output, err) = p.communicate()
                                output = output.decode('utf-8')
                                success = 1
                            except:
                                success = 0
                                pass

                            serviceRestartResult = {
                                "id" : jobsdict['id'],
                                "message" : '',
                                "success" : success,
                            }
                            
                            self.job_results.append(serviceRestartResult)   
                            del serviceRestartResult
                        elif jobsdict['type'] == 'start_service': 
                            
                            try:
                                restartServiceDict=json.loads(jobsdict['arguments'])

                                serviceName = restartServiceDict['service_name'] + ".service"
                                
                                p =  subprocess.Popen(["systemctl", "start", serviceName], stdout=subprocess.PIPE)
                                (output, err) = p.communicate()
                                output = output.decode('utf-8')
                                success = 1
                            except:
                                success = 0
                                pass

                            serviceRestartResult = {
                                "id" : jobsdict['id'],
                                "message" : '',
                                "success" : success,
                            }
                            
                            self.job_results.append(serviceRestartResult)   
                            del serviceRestartResult

                        elif jobsdict['type'] == 'get_network_interfaces':      
                            network_interfaces = []
                            try:
                            
                                stats   = psutil.net_if_stats()
                                bandw   = psutil.net_io_counters(pernic=True)
                                                                
                                for nic, addrs in psutil.net_if_addrs().items():
                                                                           
                                    network = {
                                        "name" : nic,
                                        "speed" : '',
                                        "mac_address" : '',
                                        "link" : '',
                                        "type" : '',
                                        "bytes_sent" : '',
                                        "bytes_received" : ''
                                    }

                                    gateway_addresses = []

                                    gateway_info = self.get_default_gateway()

                                    if (gateway_info["interface"] == nic):
                                        gateway_addresses.append(gateway_info["gateway"])
                                        network["gateway_addresses"] = gateway_addresses
                                            
                                    if nic in stats:
                                        st = stats[nic]
                                        
                                        network["speed"] = st.speed * 1000000
                                        if (st.isup):
                                            network["link"] = 1
                                        else:
                                            network["link"] = 0
                                    
                                    if nic in bandw:
                                        bw = bandw[nic]
                                        
                                        network["bytes_sent"] = bw.bytes_sent
                                        network["bytes_received"] = bw.bytes_recv
                                    
                                    ipAddress = []
                                    ip_info = []

                                    for addr in addrs:
                                        if (addr.family == socket.AF_INET or addr.family == socket.AF_INET6):
                                            ipAddress.append(addr.address)
                                            ip_temp = {
                                                "ip_address" : '',
                                                "subnet" : ''
                                            }
                                            ip_temp["ip_address"] = addr.address
                                            ip_temp["subnet"] = addr.netmask
                                            ip_info.append(ip_temp)
                                            del ip_temp

                                            
                                        if (addr.family == psutil.AF_LINK):
                                            network["mac_address"] = addr.address
                                        
                                    
                                    network["ip_addresses"] = ipAddress
                                    network["ip_info"] = ip_info
                                    
                                            
                                    network_interfaces.append(network)
                                    del network
                            except:
                                pass    
                            
                            networkResult = {
                                "id" : jobsdict['id'],
                                "message" : '',
                                "success" : 1,
                                "network_interfaces" : network_interfaces
                            
                            }
                            
                            self.job_results.append(networkResult)   
                            del networkResult                        
                        
                        elif jobsdict['type'] == 'get_storage_devices':      
                            storage_drives = []
                            try:
                                                    
                                for drive in psutil.disk_partitions():
                                    free = '';
                                    total = '';
                                    try:
                                        stat= os.statvfs(drive.mountpoint)
                                            # use f_bfree for superuser, or f_bavail if filesystem
                                            # has reserved space for superuser
                                        free = (stat.f_bavail * stat.f_frsize)
                                        total = (stat.f_blocks * stat.f_frsize)
                                    except:
                                        pass
                                    
                                    storage = {
                                        "device" : drive.device,
                                        "mount_point" : drive.mountpoint,
                                        "file_system" : drive.fstype,
                                        "options" : drive.opts,
                                        "size" : total,
                                        "free" : free
                                    }
                                    
                                    
                                    storage_drives.append(storage)
                                    del storage
                            except:
                                pass    
                            
                            storageResult = {
                                "id" : jobsdict['id'],
                                "message" : '',
                                "success" : 1,
                                "storage_devices" : storage_drives
                            
                            }
                            
                            self.job_results.append(storageResult)   
                            del storageResult
                        
                        elif jobsdict['type'] == 'kill_process':
                            killdict=json.loads(jobsdict['arguments'])
                            
                            try:
                                pid = int(killdict['pid'])
                                p = psutil.Process(pid)
                                p.terminate()  #or p.kill()
                                success = 1
                            except:
                                success = 0
                                pass
                            
                            
                            processResult = {
                                "id" : jobsdict['id'],
                                "message" : '',
                                "success" : success
                            
                            }
                            
                            self.job_results.append(processResult)   
                            del processResult
                            
                        elif jobsdict['type'] == 'restart':
                            os.system('reboot now')

                            restartResult = {
                                "id" : jobsdict['id'],
                                "message" : '',
                                "success" : 1
                            
                            }
                            
                            self.job_results.append(restartResult)   
                            del restartResult
                            
                        elif jobsdict['type'] == 'start_service':
                            startDict=json.loads(jobsdict['arguments'])
                            
                            try:
                            
                                os.system('service ' + startDict['service_name'] + ' start')
                                success = 1

                            except:
                                success = 0
                                pass

                            startServiceResult = {
                                "id" : jobsdict['id'],
                                "message" : '',
                                "success" : success
                            
                            }
                            
                            self.job_results.append(startServiceResult)   
                            del startServiceResult

                        elif jobsdict['type'] == 'stop_service':
                            stopDict=json.loads(jobsdict['arguments'])
                            
                            try:
                            
                                os.system('service ' + stopDict['service_name'] + ' stop')
                                success = 1

                            except:
                                success = 0
                                pass

                            stopServiceResult = {
                                "id" : jobsdict['id'],
                                "message" : '',
                                "success" : success
                            
                            }
                            
                            self.job_results.append(stopServiceResult)   
                            del stopServiceResult

                        elif jobsdict['type'] == 'ping_ip_addresses':
                            pingCommand=json.loads(jobsdict['arguments'])

                            ping_results = []
                            success = 0

                            try:
                                success = 1             

                                for ip_address in pingCommand["ip_addresses"]:
                                    pingResult = 0
                                    dns_name = ''
                                    message = ''
                                    mac_address = ''
                                    ports = []
                                    
                                    try:

                                        Pingparam = '-n' if platform.system().lower()=='windows' else '-c'

                                        # Building the command. Ex: "ping -c 1 google.com"
                                        commandPar = ['ping', Pingparam, '1', ip_address]

                                        pingResultTemp = subprocess.call(commandPar)

                                        if pingResultTemp == 0:
                                            pingResult = 1

                                        mac_address = self.getmacbyip(ip_address, False)
                
                                    except:
                                        pass

                                    try:
                                        dns_name,alias,addresslist = socket.gethostbyaddr(ip_address) 
                                    except:
                                        pass

                                    try:

                                        n = requests.get('http://' + ip_address + '/', timeout=1.000)

                                        port = {
                                            "response" : 1,
                                            "result" : n.text,
                                            "application" : 'http',
                                            "protocol": 'tcp',
                                            "port" : '80',
                                            "version" : n.headers["Server"]
                                        }

                                        regex = re.compile('<title>(.*?)</title>', re.IGNORECASE|re.DOTALL)
                                        message = regex.search(n.text).group(1)

                                        ports.append(port)
                                        del port
                                        del n

                                    except:
                                        pass

                                    try:
                                        
                                        n = requests.get('https://' + ip_address + '/', timeout=1.000, verify=False)

                                        port = {
                                            "response" : 1,
                                            "result" : n.text,
                                            "application" : 'https',
                                            "protocol": 'tcp',
                                            "port" : '443',
                                            "version" : n.headers["Server"]
                                        }

                                        #regex = re.compile('<title>(.*?)</title>', re.IGNORECASE|re.DOTALL)
                                        #message = regex.search(n.text).group(1)

                                        ports.append(port)
                                        del port
                                        del n

                                    except:
                                        pass
                                    
                                    pingReturn = {
                                        "ip_address" : ip_address,
                                        "response" : pingResult,
                                        "dns_name" : dns_name,
                                        "round_trip_time" : '',
                                        "message" : message,
                                        "mac_address" : mac_address,
                                        "ports" : ports
                                    }
                    
                                    ping_results.append(pingReturn)
                                    del pingReturn
                                    del ports
                            except:
                                pass  

                            pingResult = {
                                "id" : jobsdict['id'],
                                "message" : '',
                                "success" : success,
                                "ping_results" : ping_results
                            
                            }
                            
                            self.job_results.append(pingResult)   
                            del pingResult

                        elif jobsdict['type'] == 'get_local_users':

                            local_users_results = []
                            local_groups_results = []
                            
                            success = 0

                            try: 
                                os_type = sys.platform.lower()

                                if "linux" in os_type:
                                                    
                                    try:

                                        import pwd, grp
                            
                                        for p in pwd.getpwall():

                                            temp_user = {
                                                "name" : p.pw_name,
                                                "sid" : str(p.pw_uid),
                                                "full_name" : p.pw_gecos,
                                                "enabled" : True,
                                                "group_members" : [
                                                    {
                                                        "sid" : str(p.pw_gid)
                                                    }
                                                ]
                                            }

                                            local_users_results.append(temp_user)
                                            del temp_user

                                        for g in grp.getgrall():

                                            temp_group = {
                                                "name" : g.gr_name,
                                                "sid" : str(g.gr_gid)
                                            } 

                                            local_groups_results.append(temp_group)
                                            del temp_group

                                        success = 1

                                    except:
                                        pass
                                
                               
                            except:
                                pass

                            usersResult = {
                                "id" : jobsdict['id'],
                                "message" : '',
                                "success" : success,
                                "local_users" : local_users_results,
                                "local_groups" : local_groups_results
                            
                            }

                            self.job_results.append(usersResult)   
                            del usersResult

                        else:
                            #unknown command
                            #print('Unknown task')
                            commandResult = {
                                "id" : jobsdict['id'],
                                "message" : 'Unknown task ' + jobsdict['type'],
                                "success" : 0
                            
                            }
                            
                            self.job_results.append(commandResult)
            if 'checkin_interval' in returned:
                self.checkin_interval = int(returned['checkin_interval'])
            
        else:
            print('Unsuccessful checkin')
            print(returned)

    def post(self, url, values, verifyHTTPS = True):
        '''POSTs to MSP Server API'''
        try:
            response = requests.post(url, json = values, verify=verifyHTTPS, timeout=20)
            if response.status_code != 200 or not response.text:
                print(response)
                print(response.text)
                return []
            else:
                try:
                    data = response.json()
                    return data
                except:
                    print('JSON Decoding failed: ' + response.text)
                    return []
        except Exception as error:
            print(error)
            return []
                