#!/usr/bin/env python3

# Modules written for this agent
from datagatherer import DataGatherer
from apimanager import APIManager

# System modules for daemonisation
import signal
from threading import Event
from threading import Thread
import time
import os
import sys

clientVersion = "2.0.1"

# Other modules
import configparser

# This can be removed once all have been migrated from 1.2 to 1.3
import json

class BTAgent():

    kill_now = False

    def __init__(self):
        # Set signals for termination from supervisor
        signal.signal(signal.SIGINT, self.exit_gracefully)
        signal.signal(signal.SIGTERM, self.exit_gracefully)

        # Create thread event to sleep while responding to a signal
        self.exit = Event()

        #Start in correct folder (shouldn't be required)
        #os.chdir('/opt/msp-agent/')

        # Get config
        self.config = configparser.ConfigParser()

        # get config
        if not os.path.exists('config.ini'):
            self.install()
        self.config.read('config.ini')

        # Singletons for getting and posting data
        self.dg = DataGatherer(clientVersion)
        self.api = APIManager(self.config['API'])
        
        
    def install(self):
        # create temporary API manager to test urls
        api = APIManager()
        dg = DataGatherer()

        # migrate json config to configparser
        if os.path.exists('config.json'):
            jsonConfig = json.load(open('config.json'))
            self.config['API'] = {
                'url': jsonConfig['url'],
                'uuid': jsonConfig['uuid']
            }
            self.config['Agent'] = {
                'id': jsonConfig['id'],
                'version': clientVersion
            }

            # test that the url is still valid
            if api.test(self.config['API']['url']):
                with open('config.ini', 'w') as configfile:
                    self.config.write(configfile)
                os.remove('config.json')
                return

        url = ''
        testSuccess = False

        while not testSuccess:
            # Set URL of Bluetrait server API
            url = input('Bluetrait MSP API URL: ')
            if api.test(url):
                testSuccess = True
            else:
                print('URL test failed.')

        created = api.create(url, dg.hostname())

        if 'success' in created and created['success'] == 1:
            self.config['API'] = {
                'url': url,
                'uuid': created['uuid'],
                }
            self.config['Agent'] = {
                'id': created['id'],
                'version': clientVersion
            }

            with open('config.ini', 'w') as configfile:
                self.config.write(configfile)
        else:
            print('Installing agent failed.')
            sys.exit()


    def exit_gracefully(self, signum, frame):
        self.kill_now = True
        self.exit.set()
    
    def timer(self):

        try:
            # dg.gather() does not currently respond to interrupts
            #get_jobs added to datagatherer argument -LVT
            self.dg.gather(['name', 'agent_version', 'ipv4_address', 'ipv6_address', 'wan_ipv4_address', 'wan_ipv6_address', 'primary_drive', 'primary_drive_size', 'primary_drive_free_space', 'uptime', 'load_average', 'last_sw_update', 'ram', 'cpu', 'os_version', 'os_name', 'os_build', 'agent_type', 'other_data', 'get_jobs', 'free_ram', 'hw_serial_number', 'job_results'], 25)
            #print statement removed to prevent output file becoming too large
            #print(self.dg.data)
            
            self.api.checkin(self.dg.data)
        except:
            pass


    def loop(self):
        # True if SIGINT or SIGTERM recieved
        while not self.kill_now:
            
            thread = Thread(target=self.timer)
            thread.start()

            self.exit.wait(30)
